#!/usr/bin/env python
# -*- coding: utf-8 -*-

import Queue
import logging
from threading import Event, Thread
import time

from SerialRxTx import SerialRxTx
from CommandParser import CommandParser

_logger = logging.getLogger(__name__)
ENABLE_DEBUG_LOG = True
ENABLE_RX_DEBUG_LOG = False
ENABLE_TX_DEBUG_LOG = False
GUI_TIMEOUT = 5.0                 # Zeit in [s], nach der die GUI als 'nicht verbunden' angenommen wird.


class Communication(object):
    """!
    Kommunikation mit der Bedieneinheit
    """
    def __init__(self, gs1controller):
        """!
        Initialisierung der Kommunikation. Öffnet die Serielle Schnittstelle und startet den Thread.
        """
        _logger.debug("Communication init")
        self._name = "Communication"
        self._gs1Controller = gs1controller
        self._command_parser = CommandParser(self._gs1Controller)
        # Serielle Kommunikation
        self._serRxTx = SerialRxTx('/dev/ttyO1')    # Bei GS1 mit ttyO1 verbinden
        # --- buffer ---
        self._input_buffer = Queue.Queue()
        self._output_buffer = Queue.Queue()
        self._gui_available = False
        self._sendingFileBlocks = False
        self._prepareForFileBlocks = False
        # --- thread ---
        self._process_time = 0.05
        self._stopped = Event()
        self._stopped.clear()
        self._thread_communication = Thread(target=self._process, name=self._name)
        self._thread_communication.setDaemon(True)
        self._thread_communication.start()
        _logger.debug("Communication start")

    def exit(self):
        """!
        Beendet den Thread.
        """
        self._stopped.set()
        time.sleep(self._process_time)
        self._serRxTx.exit()
        self._serRxTx = None

    def write(self, value):
        """!
        Mit dieser Funktion sendet der Gs1Controller ein Kommando an die Bedieneinheit
        @param value: Kommando das gesendet wird
        """
        # prevent filling buffer if prepared for file upload
        if self._sendingFileBlocks or not self._prepareForFileBlocks:
            self._input_buffer.put(value)

    def read(self):
        """!
        Mit dieser Funktion holt der Gs1Controller die von der Bedieneinheit empfangenen Kommandos ab
        @return Kommandos, die empfangen wurden
        """
        if self._output_buffer.empty():
            return None
        else:
            rc = self._output_buffer.get()
            self._output_buffer.task_done()
            return rc

    @property
    def gui_available(self):
        """!
        Gibt True zurück, wenn die GUI mit der GS1 verbunden ist
        @return True, wenn die GUI mit der GS1 verbunden ist, sonst False
        """
        return self._gui_available

    def _parse_message(self, message):
        """!
        Trennt die Botschaft in Befehle auf und legt diese in die Ausgangs-Queue.
        @param message: GUI Botschaft als String mit '$$' als Trennzeichen.
        @return: True, wenn ein 'ack' als Bestätigung empfangen wurde
        """

        retval = False

        try:
            commandstrings = message.split('$$')        # Botschaft in Kommandos trennen
            commandstrings = commandstrings[:-1]        # Checksumme nicht betrachten

            for commandstring in commandstrings:        # Kommandos nacheinander ansehen
                if commandstring == "nack":             # GUI meldet fehlerhaftes Kommando
                    _logger.error("received 'nack' in message: {}".format(message))
                elif commandstring == "ack":            # GUI meldet Alles prima
                    retval = True
                else:
                    self._command_parser.executeCommand(commandstring)      # Kommando sofort ausführen

        except:     # TODO: Exeption Type
            _logger.error("error when parsing message: {}".format(message))

        return retval

    def _process(self):
        """!
        Holt die empfangenen Daten via GS1 Controller ab und Verarbeitet die empfangenen Daten
        """
        gui_response_timeout = time.time() + GUI_TIMEOUT    # Timeout für Antwort der GUI

        while not self._stopped.wait(self._process_time):
            # Von serieller Schnittstelle lesen und ggf. in MessageParser geben.
            if self._serRxTx != None:
                rxmessage = self._serRxTx.read()
                if rxmessage:
                    gui_response_timeout = time.time() + GUI_TIMEOUT  # Timeout für Antwort der GUI
                    self._gui_available = True
                    if ENABLE_RX_DEBUG_LOG:
                        _logger.debug("rec: {}".format(rxmessage))
                    self._parse_message(rxmessage)

                if self._gs1Controller.get_gui_connection_detection() and gui_response_timeout < time.time():
                    self._gui_available = False
                    print ("reset der seriellen Kommunikation erfolgreich: {}".format("True" if (self._reset_serial_com()) else "False"))

                if self._sendingFileBlocks:
                    txblock = self._input_buffer.get()
                    self._serRxTx.write(txblock)
                else:
                    # Kommandos von GS1Controller zu Botschaft zusammensetzen und versenden
                    command_list = []
                    endmessage = 'ack'
                    while not self._input_buffer.empty():
                        txcommand = self._input_buffer.get()
                        if txcommand == 'nack':
                            if ENABLE_DEBUG_LOG:
                                _logger.debug("send txmessage with 'nack'")
                            endmessage = 'nack'
                        else:
                            command_list.append(txcommand)
                        self._input_buffer.task_done()

                    if command_list != []:
                        command_list.append(endmessage)
                        txmessage = '$$'.join(command_list)
                        self._serRxTx.write(txmessage)
                        if ENABLE_TX_DEBUG_LOG:
                            _logger.debug("send: {}".format(txmessage))

            time.sleep(self._process_time)

    def enable_gui_detection(self, enable):
        """!
        Schaltet die GUI-Erkennung aus für SW-Update
        @param enable: Erkennung an/aus True/False
        """
        self._gs1Controller.enable_gui_connected_detection(enable)

    def _reset_serial_com(self):
        """!
        Setzt die serielle Schnittstelle neu auf
        """
        try:
            self._serRxTx.exit()
            self._serRxTx = None
            self._serRxTx = SerialRxTx('/dev/ttyO1')    # Bei GS1 mit ttyO1 verbinden
            return True
        except:
            return False
            
    def prepare_for_file_blocks(self, val):
        if val != False and val != True:
            return False

        self._prepareForFileBlocks = val
        return True

    def set_sending_file_blocks(self, val):
        if val != False and val != True:
            return False

        self._sendingFileBlocks = val
        self._serRxTx.set_sending_file_blocks(val)

        if val == False:
            self._prepareForFileBlocks = False

        return True
